#!/bin/bash
#
# Copyright (c) 2012- Fixpoint Inc. All rights reserved.
# ---
# Kompira簡易インストーラ：
#   - 1台構成版
#   - Apache/PostgreSQLを使用
#
# 動作要件：
#   - x86_64
#   - CentOS 8.x/7.x系 (8.x 推奨)
#   - Python 3.6系
#
# インストールに必要な追加レポジトリ:
#                       : CentOS8/RHEL8                      : CentOS7                  : RHEL7
# ----------------------------------------------------------------------------------------------------------------------
# python3/python3-devel : なし                               : centos-release-scl-rh    : rhel-7-server-rpms/rhel-7-server-optional-rpms
# (*1)                                                                                    (rhui-)rhel-7-server-rhui-rpms/(rhui-)rhel-7-server-rhui-optional-rpms
# postgresql12-server   : pgdg-redhat-repo-latest.noarch.rpm : 同左                     : 同左
# postgresql12-contrib
#
# *1 RHUIのレポジトリの問題で、Python3のインストールに失敗する場合、--rhel-option-repoオプション引数で、python3-develの含まれるレポジトリを指定するか、
#    それでも上手くいかない場合は、事前に手動でpython3/python3-develをインストールしておき、--skip-python3-install オプションを指定してインストールする。
#
shopt -s extglob
SETUP_TYPE="Install the Kompira"
INSTALL_LOG="install.$$.log"
THIS_DIR=$(dirname $(readlink -f $0))

. $THIS_DIR/scripts/setup_utils.sh

KOMPIRA_LOGIN=/bin/bash
KOMPIRA_PKGNAME_DEFAULT=Kompira
KOMPIRA_PKGNAME_JOBMNGR=Kompira-jobmngr
KOMPIRA_PKGNAME_SENDEVT=Kompira-sendevt
KOMPIRA_WHLFILE_DEFAULT="$THIS_DIR/dist/Kompira-$VERSION-py3-none-any.whl"
KOMPIRA_WHLFILE_JOBMNGR="$THIS_DIR/dist/Kompira_jobmngr-$VERSION-py3-none-any.whl"
KOMPIRA_WHLFILE_SENDEVT="$THIS_DIR/dist/Kompira_sendevt-$VERSION-py3-none-any.whl"
KOMPIRA_TEST_URL="https://localhost/"
KOMPIRA_DB_TABLE_EXISTS=false
KOMPIRA_DB_USER_EXISTS=false
KOMPIRA_EXTRA_CONFIGS="kompira_audit.yaml"
#
# DB暗号化フィールド用共有秘密鍵
#
DB_SECRET_KEY_LEN=12        # 自動作成時のキー長
DB_SECRET_KEY_MIN_LEN=8     # kompira_site.parameters.KOMPIRA_SECRET_KEY_MIN_LEN に合わせる
DB_SECRET_KEY=
#
# HTTP(S) アクセス可能かチェックする URL
#
CHECK_ACCESSIBILITY="https://bootstrap.pypa.io https://download.postgresql.org https://dl.fedoraproject.org"

#
# SCL レポジトリ情報
#
SCL_REPOPKG="centos-release-scl-rh"
SCL_YUMREPO="http://mirror.centos.org/centos/${RHEL_VERSION}/extras/x86_64/Packages/centos-release-scl-rh-2-3.el${RHEL_VERSION}.centos.noarch.rpm"

#
# RHEL レポジトリ情報(RHEL7で必要となる)
#
RHEL_OPTION_REPONAME="rhel-${RHEL_VERSION}-server-optional-rpms"
RHUI_OPTION_REPONAME="rhel-${RHEL_VERSION}-server-rhui-optional-rpms"

#
# postgres レポジトリ／パッケージ情報
#
PG_YUMREPO="https://download.postgresql.org/pub/repos/yum/reporpms/EL-${RHEL_VERSION}-${REQUIRES_ARCH}/pgdg-redhat-repo-latest.noarch.rpm"

#
# epel レポジトリ情報
#
EPEL_REPONAME="epel"
EPEL_REPOPKG="epel-release"
EPEL_YUMREPO="https://dl.fedoraproject.org/pub/epel/epel-release-latest-${RHEL_VERSION}.noarch.rpm"

#
# localhost チェック用ワンライナー
#
PYTHON_CHECK_LOCALHOST='from socket import *; [socket(*res[:3]) and exit(0) for res in getaddrinfo("localhost", 0, 0, SOCK_STREAM)] or exit(1)'

#
# firewalld/iptables の許可ポート
#
FIREWALLD_ACCEPT_PORTS="http https 5672/tcp"
IPTABLES_ACCEPT_PORTS="80/tcp 443/tcp 5672/tcp"

#
# SELinux のモード
#
SELINUX_MODE=permissive

#
# インストール自体に必要なパッケージ
#
PACKAGES_FOR_UPDATE="nss"
PACKAGES_FOR_INSTALL="yum-utils curl nss"
PACKAGES_FOR_LANGPACK="glibc-langpack-en glibc-langpack-ja"

#
# kompira_sendevt に必要なパッケージ
#
PACKAGES_FOR_SENDEVT=""
PACKAGES_FOR_SENDEVT_ONLINE=""

#
# kompira_jobmngr に必要なパッケージ
#
# gcc: pycrypto
# openssl-devel:
# krb5-devel は openssl-devel に依存してインストールされる
PACKAGES_FOR_JOBMNGR="krb5-workstation openssl libffi"
PACKAGES_FOR_JOBMNGR_ONLINE="gcc make krb5-devel openssl-devel libffi-devel"

#
# kompira サーバに必要なパッケージ
#
PACKAGES_FOR_KOMPIRA="httpd mod_ssl mercurial git sudo iproute jq memcached postfix expect"
PACKAGES_FOR_KOMPIRA_ONLINE="httpd-devel"
PACKAGES_FOR_KOMPIRA_RABBITMQ="erlang rabbitmq-server"
PACKAGES_FOR_KOMPIRA_POSTGRESQL="${PG_PREFIX}-server ${PG_PREFIX}-contrib"
#PACKAGES_FOR_KOMPIRA_POSTGRESQL_ONLINE="${PG_PREFIX}-devel"

#
# 冗長構成セットアップ(setup_cluster.sh)に必要なパッケージ
#
PACKAGES_FOR_SETUP_CLUSTER="ipcalc"

#
# 追加でインストールするパッケージ
#
: ${WITH_RPM:=""}
: ${WITH_WHL:=""}

#
# extra パッケージ構築に必要なパッケージ
#
: ${PACKAGES_FOR_EXTRA:="createrepo rpm-build"}

#
# extra パッケージに追加で加えるパッケージ
#
: ${PACKAGES_FOR_OFFLINE:=""}
: ${REQUIREMENTS_FOR_OFFLINE:=""}

#
# gdb/debuginfo パッケージ
#
: ${PACKAGES_FOR_GDB:="gdb glibc"}
: ${DEBUGINFO:="glibc"}

#
# ロケールの設定
#
: ${CONFIG_LOCALE_LANG:=false}
: ${CONFIG_LOCALE_TIMEZONE:=false}
: ${LOCALE_LANG:=""}
: ${LOCALE_TIMEZONE:="Asia/Tokyo"}
: ${LOCALE_CLOCK_UTC:="false"}

#
# LOCALE_LANG のデフォルトは postgres の locale 設定があればそれを引き継ぐ
# なければ System Locale の設定を引き継ぐ
#
if [ -z "$LOCALE_LANG" ]; then
    PGSQL_LANG=$(psql -U $PGSQL_USER -d $PGSQL_DB -At -c "show lc_collate;" 2>/dev/null)
    if [ -n "$PGSQL_LANG" ]; then
        LOCALE_LANG=$PGSQL_LANG
    else
        SYSTEM_LOCALE=$(localectl|grep 'System Locale')
        LOCALE_LANG=${SYSTEM_LOCALE#*=}
    fi
fi

#
# インストール時の proxy の設定
#
: ${TEMP_PROXY_URL:="$PROXY_URL"}
: ${TEMP_NO_PROXY:="$NO_PROXY"}

HTTPS_MODE=true
FORCE_MODE=false
BACKUP_MODE=true
BACKUP_PROCESS=false
INITDATA_MODE=false
INITFILE_MODE=false
EXTRA_MODE=false
EXTRA_WITH_HA=true
JOBMNGR_MODE=false
SENDEVT_MODE=false
KOMPIRA_SERVER=localhost
REQUIRES_ONLY=false
WITH_GDB=false
INSTALL_ONLY=false
DRY_RUN=false
INSTALL_ONLY=false
SKIP_PYTHON3_INSTALL=false
SKIP_CLUSTER_START=false

SHOW_OPTIONS="$SHOW_OPTIONS LOCALE_LANG PATH HTTPS_MODE FORCE_MODE BACKUP_MODE BACKUP_PROCESS INITDATA_MODE INITFILE_MODE OFFLINE_MODE JOBMNGR_MODE SENDEVT_MODE PROXY_URL NO_PROXY KOMPIRA_SERVER INSTALL_ONLY DRY_RUN"

check_localhost()
{
    $PLATFORM_PYTHON -c "$PYTHON_CHECK_LOCALHOST" > /dev/null 2>&1
}

check_environ()
{
    echo_title "Check current environment."
    if [ -n "$TEMP_PROXY_URL" ]; then
        # インストール時のプロキシ環境変数を設定する
        echo_debug "settings http_proxy for install: $TEMP_PROXY_URL"
        echo_debug "settings no_proxy for install: $TEMP_NO_PROXY"
        export http_proxy="$TEMP_PROXY_URL"
        export https_proxy="$TEMP_PROXY_URL"
        export HTTP_PROXY="$TEMP_PROXY_URL"
        export HTTPS_PROXY="$TEMP_PROXY_URL"
        export no_proxy="$TEMP_NO_PROXY"
    fi

    check_arch
    check_system
    check_subscription $OFFLINE_MODE
    if $EXTRA_MODE; then return 0; fi

    if ! $OFFLINE_MODE; then
        # ネットワークアクセスできるか確認する
        echo_info "Check network accessibility"
        local index=0
        for url in $CHECK_ACCESSIBILITY; do
            let index=$index+1
            verbose_run curl -x "$TEMP_PROXY_URL" --connect-timeout 10 -kvfI "$url" > /dev/null 2> $TMPDIR/test-accessibility$index.out
            if [ $? != 0 ]; then
                cat $TMPDIR/test-accessibility$index.out
                echo_error "[ERROR] Can not connect to $url"
                echo_error "Please make sure this server can connect to the internet. Or check the proxy settings."
                abort_setup
            fi
            echo_info "OK: $url"
        done
    fi

    if $OFFLINE_MODE; then
        # オフラインインストールに必要な extra パッケージが展開されているかチェック
        local dir
        if [ ! -f $KOMPIRA_EXTRA_REPO ]; then
            abort_setup "$KOMPIRA_EXTRA_REPO not found!"
        fi
        for dir in $KOMPIRA_EXTRA_PIP $KOMPIRA_EXTRA_RPM $KOMPIRA_EXTRA_WHL; do
            if [ ! -d $dir ]; then
                abort_setup "$dir not found!"
            fi
        done
    fi
    if ! $JOBMNGR_MODE && ! $SENDEVT_MODE; then
        # 旧バージョンの PostgreSQL チェック
        if rpm -q postgresql > /dev/null 2>&1; then
            abort_setup "Old postgresql is installed." \
                        "Please uninstall postgreql and re-execute."
        fi
        # localhost に対して正しいアドレス情報を返すことをチェック
        if ! check_localhost; then
            abort_setup "The address of localhost is invalid." \
                        "Please make sure localhost in set in /etc/hosts"
        fi
    fi
}

check_cluster()
{
    if $EXTRA_MODE; then return 0; fi
    echo_title "Check cluster status."
    get_cluster_status
    if $CLUSTER_RUNNING; then
        # 冗長構成が実行中の場合
        #   マスターノードではインストールを禁止する
        #   スレーブノードでは --force オプションをチェックする
        if $CLUSTER_MASTER; then
            abort_setup "Pacemaker is runnning (MASTER), so cannot install." \
                        "To install, demote to slave or stop the pacemaker."
        elif ! $FORCE_MODE; then
            abort_setup "Pacemaker is runnning (SLAVE), so cannot install." \
                        "To install, use --force option."
        fi
    fi
}

check_python()
{
    if ! $SKIP_PYTHON3_INSTALL; then
        install_python3
    fi
    echo_title "Check python version."
    local result=$($PYTHON -V 2>&1)
    case $result in
        $REQUIRES_PYTHON3)
            echo_info "Confirmed Python version: $result"
            echo_info "PYTHON=$PYTHON"
            ;;
        *)
            echo_info "$PYTHON"
            abort_setup "$REQUIRES_PYTHON3 is required"
            ;;
    esac
}

check_kompira_package()
{
    if $EXTRA_MODE; then return 0; fi
    local -x PIP=$(which pip 2> /dev/null)
    if [ -z "$PIP" ]; then
        return 0
    fi
    OLD_KOMPIRA_VERSION="$(get_kompira_version)"
    echo_info "VERSION=$OLD_KOMPIRA_VERSION [pip=$PIP]"
    case $OLD_KOMPIRA_VERSION in
        "")
            : # echo "*** Kompira はインストールされていません"
            ;;
        ${BRANCH}*)
            echo_info "A compatible version is installed."
            ;;
        *)
            abort_setup "An incompatible version is installed, so cannot update."
            ;;
    esac
}

check_update_kompira()
{
    echo_title "Check version of Kompira installed."

    # システム環境にインストールされた Kompira (obsoleted) をチェック
    check_kompira_package

    # virtualenv 環境にインストールされている Kompira をチェック
    PATH=$KOMPIRA_BIN/:$PATH check_kompira_package
}

check_update_kompira_jobmngr()
{
    local -x PATH=$KOMPIRA_BIN/:$PATH
    local -x PIP=$(which pip 2> /dev/null)
    KOMPIRA_VER=$(get_pip_info Kompira Version)
    case $KOMPIRA_VER in
        "")
            ;;
        *)
            if $FORCE_MODE; then
                echo_warn "Kompira is already installed."
            else
                abort_setup "Kompira is already installed."
            fi
    esac
}

check_update_kompira_sendevt()
{
    local -x PATH=$KOMPIRA_BIN/:$PATH
    local -x PIP=$(which pip 2> /dev/null)
    KOMPIRA_JOBMNGR_VER=$(get_pip_info Kompira_jobmngr Version)
    case $KOMPIRA_JOBMNGR_VER in
        "")
            ;;
        *)
            if $FORCE_MODE; then
                echo_warn "Kompira jobmngr is already installed."
            else
                abort_setup "Kompira jobmngr is already installed."
            fi
    esac
}

check_update_pgsql()
{
    echo_title "Check PostgreSQL verison and Kompira database existence."
    #
    # 動作中の postgresql を検出する
    # [AMZN] /usr/lib64/pgsql9x/bin/postmaster
    # [CENT] /usr/pgsql-9.x/bin/postmaster
    #
    OLD_PG_VERSION=$(readlink "/proc/$(pgrep -o postmaster)/exe" | sed -re 's|.*/pgsql-?([0-9.]+)/.*|\1|')
    if [ -n "$OLD_PG_VERSION" ]; then
        if [ "$SYSTEM" == "AMZN" ]; then
            OLD_PG_SERVICE="postgresql${OLD_PG_VERSION/./}"
        else
            OLD_PG_SERVICE="postgresql-${OLD_PG_VERSION}"
        fi
        if [ "$OLD_PG_SERVICE" != "$PG_SERVICE" ]; then
            if $CLUSTER_RUNNING; then
                abort_setup "Detected postgresql ${OLD_PG_VERSION} and Clustering is running, so cannot update." \
                            "To install, stop the pacemaker."
            fi
            echo_warn "Detected postgresql ${OLD_PG_VERSION}, so force to INITDATA_MODE"
            INITDATA_MODE=true
            return 0
        fi
    fi
    local PSQL=$(which psql 2> /dev/null)
    if [ -z "$PSQL" ]; then
        return 0
    fi
    if $PSQL -U $PGSQL_USER -d $KOMPIRA_PG_DATABASE -c \\q; then
        LANG= $PSQL -U $PGSQL_USER -d $KOMPIRA_PG_DATABASE -c \\dt > $TMPDIR/pg_tables
        KOMPIRA_DB_TABLE_EXISTS=true
    fi
    if $PSQL -U $KOMPIRA_PG_USER -d $PGSQL_DB -c \\q; then
        KOMPIRA_DB_USER_EXISTS=true
    fi
    #
    # INITDATA モードでは Kompira データベースが存在する場合、消してよいか確認する
    #
    if ! $INITDATA_MODE; then
        return 0
    fi
    # 強制モードでは既存ユーザ／データベースがあっても継続する
    if $KOMPIRA_DB_TABLE_EXISTS; then
        echo_warn "Kompira Database '$KOMPIRA_PG_DATABASE' is exist already"
    fi
    if $KOMPIRA_DB_USER_EXISTS; then
        echo_warn "Kompira DB-user '$KOMPIRA_PG_USER' is exist already"
    fi
    if $FORCE_MODE; then
        return 0
    fi
    if $KOMPIRA_DB_TABLE_EXISTS || $KOMPIRA_DB_USER_EXISTS; then
        echo
        inquire "DELETE AND CONTINUE INSTALLATION? (yes/No)" "n"
        echo $answer
        echo
        if [ $answer != "y" ]; then
            abort_setup "Abort installation."
        fi
    fi
}

check_update()
{
    if $EXTRA_MODE; then return 0; fi
    if $JOBMNGR_MODE; then
        check_update_kompira_jobmngr
    elif $SENDEVT_MODE; then
        check_update_kompira_jobmngr
        check_update_kompira_sendevt
    else
        # 実行中の pgsql のバージョンおよび既存データベースをチェックする
        check_update_pgsql
        # インストールされている Kompira の互換性チェックする
        check_update_kompira
        # Kompira がインストールされていないなら、Pacemaker がインストール済みであっても
        # シングル構成で Kompira を新規インストールできるようにする。
        if $CLUSTER_CONFIGURED && ! $CLUSTER_RUNNING && [ -z "$OLD_KOMPIRA_VERSION" ]; then
            echo_info "Since Kompira is not installed, we will set it up as a single configuration."
            CLUSTER_CONFIGURED=false
        fi
    fi
}

check_license()
{
    if $JOBMNGR_MODE || $SENDEVT_MODE; then return 0; fi
    if $CLUSTER_RUNNING; then return 0; fi
    if [ -z "$OLD_KOMPIRA_VERSION" ]; then
        return 0
    fi
    if $INITDATA_MODE || ! $KOMPIRA_DB_TABLE_EXISTS; then
        return 0
    fi
    if ! grep -qF kompira_baseobject $TMPDIR/pg_tables; then
        echo_warn "Database table is not found: kompira_baseobject"
        return 0
    fi
    # 導入されているライセンスをチェックする
    echo_title "Check installed licence of Kompira"
    verbose_run $KOMPIRA_BIN/manage.py license_info
    if [ $? != 0 ]; then
        if ! $FORCE_MODE; then
            abort_setup "The installed license is invalid."
        fi
    fi
}

check_locale_lang()
{
    local lang=(${LOCALE_LANG/./ })
    local country=${lang[0]}
    local enc=${lang[1],,}
    local enc=${enc/-/}
    if [ "$enc" != "utf8" ]; then
       abort_setup "Only UTF-8 encoding is supported for LOCALE_LANG (specified '$LOCALE_LANG')."
    fi
    for loc in $(localectl list-locales); do
        local loc_lang=(${loc/./ })
        local loc_country=${loc_lang[0]}
        local loc_enc=${loc_lang[1],,}
        if [ "$country" = "$loc_country" -a "$enc" = "$loc_enc" ]; then
            return 0
        fi
    done
    abort_setup "The specified LOCALE_LANG '$LOCALE_LANG' is not included in the system locale list. Please install the corresponding language pack if necessary."
}

yum_clean()
{
    if $SKIP_CLEAN_YUM_CACHE; then return 0; fi
    echo_title "Clean yum cache."

    # 重複パッケージを削除する
    verbose_run package-cleanup -y --cleandupes --removenewestdupes

    # yum check-upate がキャッシュを参照しないようにクリアしておく
    verbose_run yum clean all > /dev/null

    if ! $OFFLINE_MODE; then
        echo_info "Check yum update."
        verbose_run yum -q check-update > /dev/null
        if [ "$?" -eq 1 ]; then
            echo_error "[ERROR] Can not connect to yum repository."
            echo_error "Please make sure this server can connect to the internet. Or check the proxy settings."
            abort_setup
        fi
    fi
}

configure_locale_lang()
{
    if ! $CONFIG_LOCALE_LANG; then return 0; fi
    if $EXTRA_MODE; then return 0; fi

    echo_title "Check locale language settings"
    if LANG= localectl status | grep -q -F "Locale: LANG=$LOCALE_LANG"; then
        echo_info "Confirmed language: $LOCALE_LANG"
    else
        echo_info "Change language setting ($LANG -> $LOCALE_LANG)"
        verbose_run localectl set-locale LANG=$LOCALE_LANG
        LANG=$LOCALE_LANG
    fi
}

configure_locale_timezone()
{
    if ! $CONFIG_LOCALE_TIMEZONE; then return 0; fi
    if $EXTRA_MODE; then return 0; fi

    echo_title "Check locale timezone settings"
    if LANG= timedatectl status | grep -q -F "Time zone: $LOCALE_TIMEZONE"; then
        echo_info "Confirmed timezone: $LOCALE_TIMEZONE"
    else
        echo_info "Change timezone settings ($LOCALE_TIMEZONE)"
        verbose_run timedatectl set-timezone $LOCALE_TIMEZONE
    fi
}

create_user()
{
    echo_title "Create Kompira account: $KOMPIRA_USER"
    if ! id $KOMPIRA_USER > /dev/null 2>&1; then
        verbose_run useradd -s $KOMPIRA_LOGIN $KOMPIRA_USER
    else
        verbose_run usermod -s $KOMPIRA_LOGIN $KOMPIRA_USER
    fi
    #
    # .bash_profile に virtualenv を activate する設定を追加
    # MEMO: チルダ展開の前にパラメータ展開するため eval で展開
    #
    local BASH_PROFILE=$(eval "echo ~$KOMPIRA_USER/.bash_profile")
    cp $BASH_PROFILE $TMPDIR/bash_profile

    if grep -E -i -q "^export http?_proxy=" $TMPDIR/bash_profile; then
        verbose_run \
            sed -i -r \
            -e "s|^(export https?_proxy)=.*|\1='$PROXY_URL'|i" \
            -e "s|^(export no_proxy)=.*|\1='$NO_PROXY'|" \
            $TMPDIR/bash_profile
    else
        cat >> $TMPDIR/bash_profile <<EOF

# proxy settings
export http_proxy='$PROXY_URL'
export https_proxy='$PROXY_URL'
export HTTP_PROXY='$PROXY_URL'
export HTTPS_PROXY='$PROXY_URL'
export no_proxy='$NO_PROXY'
EOF
    fi
    if ! grep -q -F "source $KOMPIRA_BIN/activate" $TMPDIR/bash_profile; then
        cat >> $TMPDIR/bash_profile <<EOF

# activate virtualenv
if [ -f $KOMPIRA_BIN/activate ]; then
    source $KOMPIRA_BIN/activate
fi
EOF
    fi
    diff_install -o $KOMPIRA_USER -g $KOMPIRA_GROUP $TMPDIR/bash_profile $BASH_PROFILE
}

create_home()
{
    echo_title "Create Kompira directory: $KOMPIRA_HOME"
    verbose_run $INSTALL -m 755 -d $KOMPIRA_HOME
}

create_virtualenv()
{
    echo_title "Create virtualenv: $KOMPIRA_ENV"
    #
    # python3 の venv 作成(pip も合わせてインストールされる)
    #
    local venv_options=""
    $PYTHON -m venv $venv_options $KOMPIRA_ENV
    exit_if_failed "$?" "Failed to create virtualenv"
    # virtualenv の pip 用 proxy 設定
    echo -e "[global]\nproxy=$PROXY_URL" > $TMPDIR/kompira_pip.conf
    if ! diff -N -u $KOMPIRA_ENV/pip.conf $TMPDIR/kompira_pip.conf; then
        verbose_run cp -S .old -b $TMPDIR/kompira_pip.conf $KOMPIRA_ENV/pip.conf
    fi
}

activate_virtualenv()
{
    source $KOMPIRA_ENV/bin/activate
}

deactivate_virtualenv()
{
    deactivate
}

install_python3()
{
    echo_title "Install Python 3 package"
    local python_packages="python3 python3-devel"
    local scl_python_mode=false
    case $SYSTEM_NAME in
        "amzn")
            if ! $OFFLINE_MODE; then
                echo_info "Install SCL repositories"
                install_repo_from_rpm "$SCL_REPOPKG" "${SCL_REPOPKG}*" $SCL_YUMREPO
            fi
            python_packages="rh-python36-python rh-python36-python-devel"
            scl_python_mode=true
            ;;
        "rhel7")
            if [ -n "${OPTION_REPONAME}" ]; then
                YUMOPT_PYTHON="--enablerepo=${OPTION_REPONAME}"
            else
                #
                # オプションレポジトリが見つからない場合は、SCLレポジトリからインストールする
                #
                echo_info "Install SCL repositories"
                install_repo_from_rpm "$SCL_REPOPKG" "${SCL_REPOPKG}*" $SCL_YUMREPO
                python_packages="rh-python36-python rh-python36-python-devel"
                scl_python_mode=true
            fi
            ;;
        "cent8" | "rhel8" | "cent7")
            ;;
    esac
    YUM_OPTION=$YUMOPT_PYTHON yum_install $python_packages
    if $scl_python_mode; then
        PYTHON=$(scl enable rh-python36 -- which python3)
    else
        PYTHON=$((which python36 || which python3.6 || which python3 || which python) 2>/dev/null)
    fi
    DEBUGINFO="$DEBUGINFO python3"
    if $EXTRA_MODE; then
        PACKAGES_FOR_OFFLINE="$PACKAGES_FOR_OFFLINE $python_packages"
    fi
    if [ ! -x $PYTHON ]; then
        abort_setup "$PYTHON not found!"
    fi
}

install_requires_prepare()
{
    echo_title "Install RPM packages for preparation"
    local yum_option=$YUMOPT_INSTALL
    # 事前にアップデートが必要なパッケージはアップデートする
    if $OFFLINE_MODE; then
        yum_option="$yum_option --noplugins --disablerepo=* --enablerepo=$KOMPIRA_EXTRA_NAME"
    fi
    verbose_run yum $yum_option -y update $PACKAGES_FOR_UPDATE
    YUM_OPTION=$YUMOPT_INSTALL yum_install $PACKAGES_FOR_INSTALL
}

install_requires_sendevt()
{
    echo_title "Install RPM packages for Kompira-sendevt"
    local packages_for_sendevt=$PACKAGES_FOR_SENDEVT
    local yum_options=$YUMOPT_SENDEVT
    if ! $OFFLINE_MODE; then
        packages_for_sendevt="$packages_for_sendevt $PACKAGES_FOR_SENDEVT_ONLINE"
    fi
    YUM_OPTION=$yum_options yum_install $packages_for_sendevt
}

install_requires_jobmngrd()
{
    echo_title "Install RPM packages for Kompira-jobmngr"
    #
    # オフラインモードでないときは必要なリポジトリを有効にする
    #
    local packages_for_sendevt=$PACKAGES_FOR_SENDEVT
    local packages_for_jobmngr=$PACKAGES_FOR_JOBMNGR
    local yum_options
    if ! $OFFLINE_MODE; then
        if [ $SYSTEM == "RHEL" ]; then
            verbose_run yum-config-manager --enable $OPTION_REPONAME > /dev/null
        fi
        packages_for_sendevt="$packages_for_sendevt $PACKAGES_FOR_SENDEVT_ONLINE"
        packages_for_jobmngr="$packages_for_jobmngr $PACKAGES_FOR_JOBMNGR_ONLINE"
    fi
    YUM_OPTION=$yum_options yum_install $packages_for_sendevt $packages_for_jobmngr
}

install_requires_kompira()
{
    echo_title "Install RPM packages for Kompira"
    #
    # オフラインモードでないときは必要なリポジトリをインストール／有効化する
    #
    local packages_for_sendevt=$PACKAGES_FOR_SENDEVT
    local packages_for_jobmngr=$PACKAGES_FOR_JOBMNGR
    local packages_for_kompira=$PACKAGES_FOR_KOMPIRA
    local packages_for_kompira_rabbitmq=$PACKAGES_FOR_KOMPIRA_RABBITMQ
    local packages_for_kompira_postgresql=$PACKAGES_FOR_KOMPIRA_POSTGRESQL
    local with_packages=$WITH_RPM
    local yum_options
    if ! $OFFLINE_MODE; then
        if [ $SYSTEM == "AMZN" ]; then
            echo_info "Enable Amazon Linux Extra packages (epel/postgresql12)"
            PYTHON=$PLATFORM_PYTHON verbose_run amazon-linux-extras enable epel
            #
            # PG_YUMREPO は /etc/redhat-release が必要となるため、インストールできない。
            # Amazon Linux 標準の extra パッケージから postgresql-12 をインストールする
            #
            PYTHON=$PLATFORM_PYTHON verbose_run amazon-linux-extras enable "postgresql${PG_VERSION}"
        else
            echo_info "Install Yum repositories (EPEL/PGSQL)"
            install_repo_from_rpm "$EPEL_REPOPKG" "${EPEL_REPOPKG}*" $EPEL_YUMREPO
            install_repo_from_rpm "$PG_REPOPKG" "${PG_REPOPKG0}*" $PG_YUMREPO
        fi
        verbose_run yum-config-manager --enable $EPEL_REPONAME > /dev/null
        if [ $SYSTEM == "RHEL" ]; then
            verbose_run yum-config-manager --enable $OPTION_REPONAME > /dev/null
        fi
        if [ $SYSTEM != "AMZN" ]; then
            verbose_run yum-config-manager --enable $PG_REPONAME > /dev/null
        fi
        case $SYSTEM_NAME in
            "cent7" | "cent8" | "rhel7" | "rhel8" | "amzn")
                curl -s https://packagecloud.io/install/repositories/rabbitmq/erlang/script.rpm.sh | bash
                curl -s https://packagecloud.io/install/repositories/rabbitmq/rabbitmq-server/script.rpm.sh | bash
                verbose_run yum -y module disable "postgresql"
                # mod-wsgiをコンパイルするために必要
                packages_for_kompira="$packages_for_kompira redhat-rpm-config"
                packages_for_kompira="$packages_for_kompira $PACKAGES_FOR_LANGPACK"
                ;;
        esac
        packages_for_sendevt="$packages_for_sendevt $PACKAGES_FOR_SENDEVT_ONLINE"
        packages_for_jobmngr="$packages_for_jobmngr $PACKAGES_FOR_JOBMNGR_ONLINE"
        packages_for_kompira="$packages_for_kompira $PACKAGES_FOR_KOMPIRA_ONLINE"
        packages_for_kompira_rabbitmq="$packages_for_kompira_rabbitmq $PACKAGES_FOR_KOMPIRA_RABBITMQ_ONLINE"
        packages_for_kompira_postgresql="$packages_for_kompira_postgresql $PACKAGES_FOR_KOMPIRA_POSTGRESQL_ONLINE"
    fi
    if $WITH_GDB; then
        with_packages="$with_packages $PACKAGES_FOR_GDB"
    fi

    YUM_OPTION=$yum_options yum_install $packages_for_sendevt $packages_for_jobmngr \
              $packages_for_kompira $packages_for_kompira_rabbitmq $packages_for_kompira_postgresql \
              $PACKAGES_FOR_SETUP_CLUSTER $with_packages

    if $WITH_GDB; then
        debuginfo_install $DEBUGINFO
    fi
}

stop_kompira()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Stop Kompira daemon and Kompira-jobmngr daemon"
    _service_stop kompirad
    _service_stop kompira_jobmngrd
}

uninstall_kompira_package()
{
    local -x PIP=$(which pip 2> /dev/null)
    if [ -z "$PIP" ]; then
        return 0
    fi
    echo_title "Uninstall old Kompira packages [pip=$PIP]"
    pip_uninstall_if_installed $KOMPIRA_PKGNAME_DEFAULT $KOMPIRA_PKGNAME_JOBMNGR $KOMPIRA_PKGNAME_SENDEVT
}

uninstall_kompira()
{
    #
    # localenv 環境以前の kompira をアンインストール
    #
    uninstall_kompira_package

    #
    # localenv 環境対応の kompira をアンインストール
    #
    local -x PATH=$KOMPIRA_BIN:$PATH
    uninstall_kompira_package
}

uninstall_packages()
{
    # 古い erlang/rabbitmq-server をアンインストール
    if ! $SENDEVT_MODE; then
        echo_title "Uninstall the old Erlang/RabbitMQ if it's installed."
        # Ferora Project の Erlang がインストールされていればアンインストールする
        if rpm -qa --qf "%{VENDOR}: %{NAME}-%{VERSION}-%{RELEASE}\n" "erlang*" | grep -F "Fedora Project:"; then
            echo_info "Uninstall old erlang packages"
            verbose_run yum $YUM_OPTION -y erase "erlang*"
            exit_if_failed "$?" "Failed to uninstall erlang"
        fi
    fi
}

install_whl_package()
{
    local pkg_file=$1
    local pkg_name=$(basename $pkg_file)
    local pip_options
    shift
    if $OFFLINE_MODE; then
        pip_options="--no-index --find-links=$KOMPIRA_EXTRA_WHL"
    fi
    if [ -n "$TEMP_PROXY_URL" ]; then
        pip_options="$pip_options --proxy=$TEMP_PROXY_URL"
    fi
    #
    # sudo -u kompira pip が展開した whl ファイルにアクセスできないので一旦 /tmp にコピーする
    # pip install でインストールするパッケージのビルト時に必要になる場合があるので PATH 環境変数を渡す
    #
    local user_tmpdir=$(sudo -u $KOMPIRA_USER mktemp -d)
    verbose_run cp $pkg_file $user_tmpdir/$pkg_name
    verbose_run $KOMPIRA_BIN/pip install $pip_options -U pip wheel setuptools
    (cd / && PATH=$PATH verbose_run $KOMPIRA_BIN/pip install $pip_options $user_tmpdir/$pkg_name $@)
    local rc=$?
    verbose_run rm -rf $user_tmpdir
    return $rc
}

install_package_sendevt()
{
    echo_title "Install Kompira-sendevt package"
    install_whl_package $KOMPIRA_WHLFILE_SENDEVT
    exit_if_failed "$?" "Failed to install Kompira-sendevt package"
}

install_package_jobmngr()
{
    echo_title "Install Kompira-jobmngr package"
    install_whl_package $KOMPIRA_WHLFILE_JOBMNGR
    exit_if_failed "$?" "Failed to install Kompira-jobmngr package"
}

install_package_default()
{
    echo_title "Install Kompira package"
    #
    # LANG=C/C.UTF-8 の場合、パッケージインストール時に失敗する可能性があるので LOCALE_LANG の値をセットしておく
    # (cf. https://github.com/fixpoint/kompira-v16/issues/1601)
    #
    LANG=$LOCALE_LANG install_whl_package $KOMPIRA_WHLFILE_DEFAULT $WITH_WHL
    exit_if_failed "$?" "Failed to install Kompira package"
    # setup_utils.sh の VERSION/BRANCH を更新する
    # TODO: 本来は whl パッケージ作成時に更新したい
    verbose_run sed -i -e "/^BRANCH=/c BRANCH=$BRANCH" -e "/^VERSION=/c VERSION=$VERSION" -e "/^SHORT_VERSION=/c SHORT_VERSION=$SHORT_VERSION" $KOMPIRA_BIN/setup_utils.sh
    # 管理コマンドの実行可能ユーザを root ユーザまたは kompira グループに所属するユーザに制限する
    verbose_run chown root:$KOMPIRA_GROUP $KOMPIRA_BIN/manage.py
    verbose_run chmod 754 $KOMPIRA_BIN/manage.py
}

backup_kompira_database()
{
    if $CLUSTER_RUNNING; then return 0; fi
    if ! $BACKUP_MODE; then return 0; fi
    if [ ! -x $KOMPIRA_BIN/manage.py ]; then
        return 0
    fi
    if [ -z "$OLD_PG_VERSION" ]; then
        return 0
    fi
    echo_title "Backup kompira database"
    if ! pgsql_has_kompira_table; then
        echo_info "Table is not exist. (SKIP)"
        return 0
    fi
    if [ -f $KOMPIRA_BACKUP ]; then
        echo_info "File exists already: $KOMPIRA_BACKUP"
        local index=1
        while [ -f "$KOMPIRA_BACKUP.$index" ]; do
            let index=$index+1
        done
        verbose_run sudo -u $KOMPIRA_USER mv $KOMPIRA_BACKUP "$KOMPIRA_BACKUP.$index"
    fi
    # loaddata 時にエラーになる contenttypes, auth.permission は除外する
    local DUMPDATA_EXCLUDES="-e contenttypes -e auth.permission -e sessions"
    # kompira1.4/django1.6 系の場合は south も除外する
    if $KOMPIRA_BIN/manage.py --help | grep -qF "[south]"; then
        DUMPDATA_EXCLUDES+=" -e south"
    fi
    # kompira 1.5.1 未満の場合は process を適切に dumpdata できないので除外する
    if [ $(ver2num $OLD_KOMPIRA_VERSION) -lt $(ver2num "1.5.1") ]; then
        DUMPDATA_EXCLUDES+=" -e kompira.process"
    elif ! $BACKUP_PROCESS; then
        DUMPDATA_EXCLUDES+=" -e kompira.process"
    fi
    verbose_run $KOMPIRA_BIN/manage.py dumpdata --traceback $DUMPDATA_EXCLUDES | su $KOMPIRA_USER -c "gzip > $KOMPIRA_BACKUP"
    exit_if_failed "$?" "Failed to backup database"
    # INITDATA_MODE のときだけ自動リストアする
    if $INITDATA_MODE; then
        verbose_run $INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_VAR_DIR/startup
        sudo -u $KOMPIRA_USER touch $KOMPIRA_VAR_DIR/startup/do_restore
    fi
}

restore_kompira_database()
{
    echo_info "Restore kompira database"
    if [ ! -f $KOMPIRA_BACKUP ]; then
        echo_warn "File not found: $KOMPIRA_BACKUP"
        return 0
    fi
    verbose_run $KOMPIRA_BIN/manage.py loaddata --ignorenonexistent $KOMPIRA_BACKUP
    exit_if_failed "$?" "Failed to restore database"
}

setup_pgsql()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure PostgreSQL settings"
    #
    # 動作中の古い postgresql を停止/無効化する
    #
    if [[ -n "$OLD_PG_SERVICE" && "$OLD_PG_SERVICE" != "$PG_SERVICE" ]]; then
        echo_info "Disable postgresql ${OLD_PG_VERSION}"
        _service_stop $OLD_PG_SERVICE
        service_disable $OLD_PG_SERVICE
    fi
    #
    # PGDATA/PG_VERSION が無い場合 initdb を実施する
    # パラメータでロケールを指定する
    #
    if [ ! -f $PG_DATADIR/PG_VERSION ]; then
        # 一度、PG_DATADIRを削除してから初期化する(initdbに失敗した残骸が残っている場合を考慮)
        verbose_run rm -rf $PG_DATADIR
        PGSETUP_INITDB_OPTIONS="--locale=$LOCALE_LANG" verbose_run $(pgsql_cmddir)/$PG_SETUP initdb
    fi
    #
    # postgresql.conf への設定追加
    #
    cp $PG_DATADIR/postgresql.conf $TMPDIR/postgresql.conf
    if ! grep -F -q "kompira.conf.d" $TMPDIR/postgresql.conf; then
        cat >> $TMPDIR/postgresql.conf <<EOF

# include kompira custom settings
include_dir 'kompira.conf.d'
EOF
        diff_install -o postgres -g postgres -m 0600 $TMPDIR/postgresql.conf $PG_DATADIR/postgresql.conf
    fi

    #
    # カスタム設定ディレクトリ・ファイルを作成する
    # [MEMO] 冗長構成用の設定ファイルはここでは作成しない
    #
    verbose_run $INSTALL -o postgres -g postgres -m 0700 -d $PG_DATADIR/kompira.conf.d
    for conffile in $THIS_DIR/config/postgresql/[0-7]*.conf; do
        local confname=$(basename $conffile)
        expand_template $conffile > $TMPDIR/pgsql-$confname
        diff_install -o postgres -g postgres -m 0600 $TMPDIR/pgsql-$confname $PG_DATADIR/kompira.conf.d/$confname
    done
    #
    # ログ出力先を作成する
    #
    verbose_run $INSTALL -o postgres -g postgres -m 0750 -d $PG_LOG_DIR
    #
    # ログローテートの設定
    #
    setup_logrotate_config "postgresql"
    #
    # postgresql サービスを開始する
    # local からのアクセスを認証なしで許可する
    # 行頭が local で始まる行中の文字列 peer を trust に置換する
    #
    service_start $PG_SERVICE
    verbose_run sed -e "/^local/{s/peer/trust/}" $PG_DATADIR/pg_hba.conf > $TMPDIR/pg_hba.conf
    diff_cp $TMPDIR/pg_hba.conf $PG_DATADIR/pg_hba.conf
    service_restart $PG_SERVICE

    #
    # INITDATA モードでは database を完全初期化する
    #
    if $INITDATA_MODE; then
        echo_info "Delete Kompira Database: $KOMPIRA_PG_DATABASE"
        verbose_run dropdb -U $PGSQL_USER --if-exists -e $KOMPIRA_PG_DATABASE
        echo_info "Delete Kompira DB-user: $KOMPIRA_PG_USER"
        verbose_run dropuser -U $PGSQL_USER --if-exists -e $KOMPIRA_PG_USER
    fi
    #
    # データベースユーザ $KOMPIRA_PG_USER を作成する
    #
    if ! psql -U $KOMPIRA_PG_USER -d $PGSQL_DB -c \\q; then
        echo_info "Create Kompira DB-user: $KOMPIRA_PG_USER"
        verbose_run createuser -U $PGSQL_USER --createdb -e $KOMPIRA_PG_USER
        exit_if_failed "$?" "Failed to create PostgreSQL user '$KOMPIRA_PG_USER'"
    fi
    #
    # データベース $KOMPIRA_PG_DATABASE を作成する
    #
    if ! psql -U $PGSQL_USER -d $KOMPIRA_PG_DATABASE -c \\q; then
        echo_info "Create Kompira Database: $KOMPIRA_PG_DATABASE"
        verbose_run createdb -U $KOMPIRA_PG_USER --encoding=utf8 -e $KOMPIRA_PG_DATABASE
        exit_if_failed "$?" "Failed to create PostgreSQL database '$KOMPIRA_PG_DATABASE'"
    fi

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ pgsql を有効化／起動したままにする
        #
        service_enable $PG_SERVICE
        if $INSTALL_ONLY; then
            service_stop $PG_SERVICE
        fi
    else
        #
        # 冗長構成のときは pgsql を停止させておく
        #
        service_stop $PG_SERVICE
    fi
}

setup_apache()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure apache settings."
    verbose_run $KOMPIRA_BIN/mod_wsgi-express install-module > $TMPDIR/mod_wsgi.conf
    exit_if_failed "$?" "Failed to install mod_wsgi"
    diff_cp $TMPDIR/mod_wsgi.conf /etc/httpd/conf.d/mod_wsgi.conf

    local httpd_conf=/etc/httpd/conf.d/kompira.conf
    local httpd_sysconf=/etc/sysconfig/httpd
    local site_packages=$($KOMPIRA_BIN/python -c "import django; print(django.__path__[0])" | sed -re 's|(.*/site-packages).*|\1|')
    local comment_if_http
    local template_prefix
    [ $HTTPS_MODE != true ] && comment_if_http="# " || comment_if_http=""
    local httpd_version=$(rpm -q httpd)
    case "$httpd_version" in
        httpd-2.4.*) template_prefix="httpd24" ;;
        *) abort_setup "Unexpected httpd version: $httpd_version" ;;
    esac

    echo_info "Append kompira configure ($template_prefix)."
    expand_template $THIS_DIR/config/httpd/$template_prefix-kompira.conf > $TMPDIR/kompira.conf
    diff_cp $TMPDIR/kompira.conf $httpd_conf

    echo_info "Change LANG setting to '$LOCALE_LANG'."
    if [ -f $httpd_sysconf ]; then
        verbose_run sed -re "s/^#?\s*(HTTPD_LANG|LANG|LC_ALL)=.*/\1=$LOCALE_LANG/" $httpd_sysconf > $TMPDIR/httpd
        if ! grep -q "^LANG=" $TMPDIR/httpd; then
            cat >> $TMPDIR/httpd <<EOF
# LANG settings
LANG=$LOCALE_LANG
EOF
        fi
        if ! grep -q "^LC_ALL=" $TMPDIR/httpd; then
            cat >> $TMPDIR/httpd <<EOF
LC_ALL=$LOCALE_LANG
EOF
        fi
    else
        # Cent/RHEL8系では、/etc/sysconfig/httpdが作成されない
        cat > $TMPDIR/httpd <<EOF
# LANG settings
LANG=$LOCALE_LANG
LC_ALL=$LOCALE_LANG
EOF
    fi
    echo_info "Change proxy settings to '$PROXY_URL'."
    if grep -E -i -q "^http?_proxy=" $TMPDIR/httpd; then
        verbose_run \
            sed -i -r \
            -e "s|^(https?_proxy)=.*|\1='$PROXY_URL'|i" \
            -e "s|^(no_proxy)=.*|\1='$NO_PROXY'|" \
            $TMPDIR/httpd
    else
        cat >> $TMPDIR/httpd <<EOF

# proxy settings
http_proxy='$PROXY_URL'
https_proxy='$PROXY_URL'
HTTP_PROXY='$PROXY_URL'
HTTPS_PROXY='$PROXY_URL'
no_proxy='$NO_PROXY'
EOF
    fi
    diff_cp $TMPDIR/httpd $httpd_sysconf

    echo_info "Append user apache to group $KOMPIRA_GROUP."
    verbose_run usermod -G $KOMPIRA_GROUP apache

    echo_info "Disable default welcome page."
    verbose_run sed -re 's/^([^#])/# \1/' /etc/httpd/conf.d/welcome.conf > $TMPDIR/welcome.conf
    diff_cp $TMPDIR/welcome.conf /etc/httpd/conf.d/welcome.conf

    echo_info "Override httpd service configue."
    cat > $TMPDIR/httpd.service <<EOF
[Service]
EnvironmentFile=/etc/sysconfig/httpd
EOF
    verbose_run mkdir -p /etc/systemd/system/httpd.service.d/
    verbose_run diff_cp $TMPDIR/httpd.service /etc/systemd/system/httpd.service.d/override.conf
    verbose_run systemctl daemon-reload

    echo_info "Override logrotate configue for httpd."
    setup_logrotate_config "httpd"

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ apache を有効化／起動する
        #
        echo_info "Enable and restart apache service"
        service_enable httpd
        if $INSTALL_ONLY; then
            #
            # mod_ssl 用の TLS 鍵初期化のために httpd-init を呼び出す
            # (デーモンではないため初期化後、サービスは終了する)
            #
            # [NOTE] この処理は --install-only モードでインストールした直後に setup_cluster.sh を
            # 実行した際、httpd リソースが起動するために必要なもの。
            #
            if [ $RHEL_VERSION == 8 ]; then
                service_start httpd-init
            fi
        else
            service_restart httpd
        fi
    fi
}

setup_rabbitmq()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure rabbitmq settings"

    if ! $CLUSTER_CONFIGURED; then
        #
        # guestユーザがlocalhost以外からもアクセスできるように設定追加
        #
        local rmq_version=$(get_rpm_version rabbitmq-server)
        if [ $(ver2num $rmq_version) -ge $(ver2num "3.8.0") ]; then
            if ! grep -q "^loopback_users" /etc/rabbitmq/rabbitmq.conf; then
                cat >> $TMPDIR/rabbitmq.conf <<EOF
loopback_users = none
EOF
                diff_cp $TMPDIR/rabbitmq.conf /etc/rabbitmq/rabbitmq.conf
            fi
        fi
        #
        # シングル構成のときだけ rabbitmq-server を有効化／起動する
        #
        service_enable rabbitmq-server
        if ! $INSTALL_ONLY; then
            service_restart rabbitmq-server
        fi
    fi
}

setup_memcached()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure memcachd settings"

    if ! $CLUSTER_CONFIGURED; then
        expand_template $THIS_DIR/config/systemd/memcached > $TMPDIR/memcached
        cat >> $TMPFILES_CONF_DIR/memcached.conf <<EOF
d      $MEMCACHED_SOCK_DIR 0755 $MEMCACHED_USER $MEMCACHED_GROUP -
EOF
        diff_cp $TMPDIR/memcached /etc/sysconfig/memcached
        verbose_run $INSTALL -o $MEMCACHED_USER -g $MEMCACHED_GROUP -m 0755 -d $MEMCACHED_SOCK_DIR
        service_enable memcached
        if ! $INSTALL_ONLY; then
            service_restart memcached
        fi
    fi
}

setup_postfix()
{
    if is_active_service postfix; then return 0; fi
    service_enable postfix
    if ! $INSTALL_ONLY; then
        service_restart postfix
    fi
}

setup_firewall()
{
    if $CLUSTER_RUNNING; then return 0; fi
    if is_active_service firewalld; then
        setup_firewalld $FIREWALLD_ACCEPT_PORTS
    elif is_active_service iptables; then
        setup_iptables $IPTABLES_ACCEPT_PORTS
    fi
}

setup_selinux()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure SELinux to $SELINUX_MODE."
    # setsebool -P httpd_can_network_connect on
    verbose_run setenforce $SELINUX_MODE
    verbose_run sed -e "/^SELINUX=/c SELINUX=$SELINUX_MODE" /etc/selinux/config > $TMPDIR/selinux
    diff_cp $TMPDIR/selinux /etc/selinux/config
}

setup_logrotate_config()
{
    # loglotate ではサフィックス .old でも設定ファイルとして読み込んでしまうので、
    # バックアップ時のサフィックスは、設定ファイルとして無視される ~ に設定する。
    local BACKUP_SUFFIX=$SIMPLE_BACKUP_SUFFIX
    export SIMPLE_BACKUP_SUFFIX="~"
    for cf in "$@"; do
        expand_template $THIS_DIR/config/logrotate/$cf > $TMPDIR/logrotate-$cf
        diff_cp $TMPDIR/logrotate-$cf /etc/logrotate.d/$cf
    done
    export SIMPLE_BACKUP_SUFFIX=$BACKUP_SUFFIX
}

setup_kompira_common()
{
    echo_title "Setup kompira common files."
    #
    # kompira グループに追加した apache アカウントで動作する httpd
    # からもログを書きこむので、パーミッションに 775 を設定する
    # 既存のログで所有者が root なものは kompira に変更する（ただし監査ログは除く）
    #
    echo_info "Create log directory: $KOMPIRA_LOG_DIR"
    verbose_run $INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_LOG_DIR
    verbose_run find $KOMPIRA_LOG_DIR -type f -user root ! -name "audit-*" -exec chown $KOMPIRA_USER:$KOMPIRA_GROUP {} \;

    echo_info "Create Kompira/Kompira-sendevt configuration file."
    expand_template $THIS_DIR/config/kompira.conf > $TMPDIR/kompira.conf
    if ! diff -N -u $KOMPIRA_HOME/kompira.conf $TMPDIR/kompira.conf; then
        verbose_run $INSTALL -S .old -b $TMPDIR/kompira.conf -m 644 $KOMPIRA_HOME/kompira.conf
    fi
}

setup_kompira_jobmngr()
{
    echo_title "Setup kompira-jobmngrd."
    local comment_if_jobmngrd_mode
    [ $JOBMNGR_MODE == true ] && comment_if_jobmngrd_mode="# " || comment_if_jobmngrd_mode=""

    echo_info "[systemd] Create kompira-jobmngrd configuration file: /etc/sysconfig/kompira_jobmngrd"
    expand_template $THIS_DIR/config/systemd/kompira_jobmngrd > $TMPDIR/kompira_jobmngrd
    diff_cp $TMPDIR/kompira_jobmngrd /etc/sysconfig/kompira_jobmngrd

    echo_info "[systemd] Enable kompira-jobmngrd service."
    expand_template $THIS_DIR/config/systemd/kompira_jobmngrd.service > $TMPDIR/kompira_jobmngrd.service
    verbose_run $INSTALL -m 664 $TMPDIR/kompira_jobmngrd.service /usr/lib/systemd/system/kompira_jobmngrd.service
    service_register kompira_jobmngrd

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ kompira_jobmngrd を有効化／起動する
        #
        echo_info "Start kompira-jobmngrd service."
        service_enable kompira_jobmngrd
        if ! $INSTALL_ONLY; then
            _service_restart kompira_jobmngrd
        fi
    fi
}

setup_kompira_datafiles()
{
    echo_title "Setup kompira data files."
    #
    # kompira グループに追加した apache アカウントで動作する httpd
    # からもアクセスするので、パーミッションに 775 を設定する
    #
    echo_info "Collect static files in kompira data directory."
    if $INITFILE_MODE; then
        verbose_run rm -rf $KOMPIRA_VAR_DIR/repository
    fi
    verbose_run $INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_VAR_DIR{,/exported,/repository,/startup,/packages,/migrations}
    verbose_run find $KOMPIRA_VAR_DIR/packages -type f ! -user kompira -name "*.json" -exec chown $KOMPIRA_USER:$KOMPIRA_GROUP {} \;
    #
    # INITDATAモード、または、キーファイルが存在しなければ作成する
    #
    if $INITDATA_MODE || ! [ -f "$DB_SECRET_KEYFILE" ]; then
        echo_info "Create secret key-file: $DB_SECRET_KEYFILE"
        if [ -z "$DB_SECRET_KEY" ]; then
            DB_SECRET_KEY=`mkpasswd -l $DB_SECRET_KEY_LEN -s 0`
        fi
        echo -n "$DB_SECRET_KEY" > "$DB_SECRET_KEYFILE"
    elif [ -n "$DB_SECRET_KEY" ]; then
        #
        # キーファイルは存在するが、secret-keyが指定されている場合、マスターキーを指定されたものに変更する
        #
        verbose_run $KOMPIRA_BIN/manage.py change_secretkey "$DB_SECRET_KEY"
    fi
    #
    # kompirad から書き込めるようにシークレットキーファイルのグループとパーミッションを変更する
    #（アップデートインストールを考慮して常に実行する）
    #
    verbose_run chgrp kompira "$DB_SECRET_KEYFILE"
    verbose_run chmod 660 "$DB_SECRET_KEYFILE"
    #
    # exported/*.dat ファイルコピー
    #
    verbose_run $INSTALL -g $KOMPIRA_GROUP -m 644 $THIS_DIR/exported/{types-*.dat,startup-*.dat,styles-*.dat,nodetypes.dat} $KOMPIRA_VAR_DIR/exported
    #
    # migrations/*.py ファイルコピー
    #
    verbose_run $INSTALL -g $KOMPIRA_GROUP -m 775 $THIS_DIR/migrations/*.py $KOMPIRA_VAR_DIR/migrations
    #
    # collectstatic はシングル、マスター、スレーブいずれでも、この段階で実施する
    #
    verbose_run $KOMPIRA_BIN/manage.py collectstatic -i PACKAGES.yaml --noinput
    #
    # pdf 形式のマニュアルをコピーする
    #
    local pdf locale
    for pdf in $THIS_DIR/Kompira-*.pdf; do
        local locale=${pdf%.pdf}
        locale=${locale##*-}
        if [ -n "$locale" ]; then
            if [ -d "$KOMPIRA_VAR_DIR/html/docs/$locale" ]; then
                $INSTALL -m 644 $pdf $KOMPIRA_VAR_DIR/html/docs/$locale/kompira.pdf
            else
                echo_warn "directory not found: $KOMPIRA_VAR_DIR/html/docs/$locale"
            fi
        fi
    done
    #
    # 次回 kompirad 起動時にデータベースの初期化を実施させるため、
    # $KOMPIRA_VAR_DIR/startup/ にファイルを作成しておく
    #
    sudo -u $KOMPIRA_USER touch $KOMPIRA_VAR_DIR/startup/{do_migrate,do_import_data,do_update_packages_info}
}

setup_kompira_settings()
{
    echo_title "Setup kompira settings."
    local -x PIP=$(which pip 2> /dev/null)
    local kompira_location=$(get_kompira_location)
    local kompira_settings="$kompira_location/kompira_site/settings.py"
    if [ ! -f $kompira_settings ]; then
        echo_warn "File not found: $kompira_settings"
        return 0
    fi
    local language_code='en-us'
    case "$LOCALE_LANG" in
        ""|c.*|C.*) ;;
        ja*) language_code="ja" ;;
        *) language_code=$(echo "${LOCALE_LANG/.*/}" | tr A-Z_ a-z-) ;;
    esac
    echo_debug "language_code=$language_code (LOCALE_LANG=$LOCALE_LANG)"
    sed -re "s/^(LANGUAGE_CODE)\s+=.*/\1 = \'$language_code\'/" \
        $kompira_settings > $TMPDIR/kompira_settings.py
    diff_cp $TMPDIR/kompira_settings.py $kompira_settings
}

setup_kompira_extra_configs()
{
    echo_title "Setup kompira extra configs."
    for cf in $KOMPIRA_EXTRA_CONFIGS; do
        if ! diff -N -u $THIS_DIR/config/$cf $KOMPIRA_HOME/$cf; then
            verbose_run $INSTALL -S .old -b $THIS_DIR/config/$cf -m 644 $KOMPIRA_HOME/$cf
        fi
    done
}

setup_kompira_daemon()
{
    echo_title "Setup kompira-daemon."
    echo_info "[systemd] Create kompira-daemon configuration file: /etc/sysconfig/kompirad"
    expand_template $THIS_DIR/config/systemd/kompirad > $TMPDIR/kompirad
    diff_cp $TMPDIR/kompirad /etc/sysconfig/kompirad

    echo_info "[systemd] Enable kompira-daemon service."
    expand_template $THIS_DIR/config/systemd/kompirad.service > $TMPDIR/kompirad.service
    verbose_run $INSTALL -m 664 $TMPDIR/kompirad.service /usr/lib/systemd/system/kompirad.service
    service_register kompirad

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ kompirad を有効化／起動する
        #
        echo_info "Start kompira-daemon service."
        service_enable kompirad
        if ! $INSTALL_ONLY; then
            _service_restart kompirad
        fi
    fi
}

setup_kompira_default()
{
    setup_kompira_datafiles
    setup_kompira_settings
    setup_kompira_extra_configs
    setup_logrotate_config "kompira_audit"
    setup_kompira_daemon
}

update_cluster()
{
    if ! $CLUSTER_CONFIGURED; then return 0; fi
    echo
    echo "*** Update clustering settings."
    echo
    #
    # アップデート時にクラスタの設定変更が必要な場合はここに実装
    #

    #
    # rsyncd サービスが enabled の場合は disable にする
    #
    if is_enabled_service rsyncd; then
        verbose_run systemctl stop rsyncd
        verbose_run systemctl disable rsyncd
    fi

    #
    # V1.6.6 以降で kompirad 起動前に pcs property を操作するための権限が必要なため
    # kompira ユーザを haclient に追加
    #
    setup_ha_user

    if $SKIP_CLUSTER_START; then return 0; fi
    #
    # pcs を起動して pacemaker の設定更新を行う
    #
    verbose_run pcs cluster start
    #
    # V1.6.6 以降では、kompirad の後で httpd を起動する必要があるため
    # res_httpd を後(vipの前)に移動する
    #
    local found_kompirad=false
    local found_httpd=false
    local before_vip=""
    local replace_httpd=false
    local webgroups=$(pcs resource group list | grep webserver:)
    for res in ${webgroups#webserver:}; do
        case $res in
            "res_httpd")
                found_httpd=true
                ;;
            "res_kompirad")
                found_kompirad=true
                ;;
            "res_vip")
                # VIPリソースがあれば、その前に移動する
                before_vip="--before res_vip"
                ;;
        esac
        if $found_httpd && ! $found_kompirad; then
            # 先に httpd が見つかった場合は res_httpd の順序を入れ替える
            replace_httpd=true
        fi
    done
    if $replace_httpd; then
        echo_info "Replace res_httpd after res_kompirad"
        verbose_run pcs resource group remove webserver res_httpd
        verbose_run pcs resource group add webserver res_httpd $before_vip
    fi
}

test_kompira()
{
    if $CLUSTER_CONFIGURED || $INSTALL_ONLY; then return 0; fi
    echo_title "Test access to kompira."
    local result=$(curl -Lksf $KOMPIRA_TEST_URL -w '%{http_code}' -o $TMPDIR/test-kompira.out)
    if [ $result != 200 ]; then
        cat $TMPDIR/test-kompira.out
        abort_setup "Access Failed!"
    fi
    echo_info "Access succeeded: $(grep "brand-version" $TMPDIR/test-kompira.out)"
}

install_kompira_sendevt()
{
    install_requires_sendevt
    if $REQUIRES_ONLY; then return 0; fi
    install_package_sendevt
    setup_kompira_common
}

install_kompira_jobmngr()
{
    install_requires_jobmngrd
    if $REQUIRES_ONLY; then return 0; fi
    install_package_jobmngr
    setup_kompira_common
    setup_kompira_jobmngr
}

install_kompira_default()
{
    install_requires_kompira
    if $REQUIRES_ONLY; then return 0; fi
    install_package_default
    setup_pgsql
    setup_apache
    setup_rabbitmq
    setup_firewall
    setup_selinux
    setup_memcached    # SELinuxをpermissiveにした後で実行する必要がある
    setup_postfix
    setup_kompira_common
    setup_kompira_jobmngr
    setup_kompira_default
    update_cluster
    test_kompira
}

install_kompira()
{
    if $SENDEVT_MODE; then
        install_kompira_sendevt
    elif $JOBMNGR_MODE; then
        install_kompira_jobmngr
    else
        install_kompira_default
    fi
}

final_check()
{
    # インストール後の最終チェック
    # 冗長構成で rabbitmq.conf がシングル用に書き換わってしまっている場合の警告 (#1870)
    local rmq_version=$(get_rpm_version rabbitmq-server)
    if $CLUSTER_CONFIGURED && [ $(ver2num $rmq_version) -ge $(ver2num "3.8.0") ]; then
        if ! grep -q "^cluster_name" /etc/rabbitmq/rabbitmq.conf; then
            echo_warn ""
            echo_warn "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
            echo_warn ""
            echo_warn "CORRUPT CONFIG DETECTED! (/etc/rabbitmq/rabbitmq.conf)"
            echo_warn ""
            echo_warn "YOU CAN RESTART THE CLUSTER AS IS."
            echo_warn "HOWEVER THE FOLLOWING MEASURES ARE RECOMMENDED."
            echo_warn ""
            local ERLANG_COOKIE=/var/lib/rabbitmq/.erlang.cookie
            if [ -f $ERLANG_COOKIE ] && [ -f $CLUSTER_CIB_FILE ]; then
                cluster_name=$(cat $ERLANG_COOKIE)
                primary_name=$(xmllint --xpath "string(//nodes/node[@id=1]/@uname)" $CLUSTER_CIB_FILE)
                secondary_name=$(xmllint --xpath "string(//nodes/node[@id=2]/@uname)" $CLUSTER_CIB_FILE)
                expand_template $THIS_DIR/config/cluster/rabbitmq.conf > $TMPDIR/rabbitmq.conf.new
                \cp -f $TMPDIR/rabbitmq.conf.new /etc/rabbitmq/rabbitmq.conf.new
                echo_warn ""
                echo_warn "Please copy /etc/rabbitmq/rabbitmq.conf.new to /etc/rabbitmq/rabbitmq.conf"
                echo_warn "before restarting the cluster."
                echo_warn ""
                echo_warn "  # cp -f -S .old -b /etc/rabbitmq/rabbitmq.conf.new /etc/rabbitmq/rabbitmq.conf"
                echo_warn ""
                echo_warn "CAUTION: PLEASE COPY ON EACH NODES WHEN BOTH CLUSTERS ARE STOPPED."
                echo_warn ""
            else
                echo_warn ""
                echo_warn "Sorry, please stop both clusters and redo cluster_setup.sh on each nodes."
                echo_warn ""
            fi
            echo_warn "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
            echo_warn ""
        fi
    fi
}

create_extra()
{
    local requirements
    local repotrack_options
    if [ -f $THIS_DIR/scripts/requires.txt ]; then
        requirements=$THIS_DIR/scripts/requires.txt
    elif [ -f $THIS_DIR/requirements/kompira.txt ]; then
        requirements=$THIS_DIR/requirements/kompira.txt
    else
        abort_setup "requirements not found!"
    fi
    eval $(grep -E "^(RHEL_HA_REPONAME|PACKAGES_FOR_HA_EL|RA_PKGURL)" $THIS_DIR/setup_cluster.sh)

    if [ $RHEL_VERSION == 7 ]; then
        PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL7
        repotrack_options="-p $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM"
        # AWS RHEL 7 の場合、repotrack にパッチをあてる必要がある
        # https://access.redhat.com/solutions/5299581
        if [ $SYSTEM == "RHEL" ] && [ -f /etc/yum/pluginconf.d/amazon-id.conf ] && grep -Fq 'init_plugins=False' /usr/bin/repotrack; then
            verbose_run sed -i.orig 's/my.doConfigSetup(fn=opts.config,init_plugins=False)/my.doConfigSetup(fn=opts.config)/g' /usr/bin/repotrack
        fi
    elif [ $RHEL_VERSION == 8 ]; then
        case $SYSTEM_RELEASEVER in
            8.[0-4]|8.[0-4].*)
                # RHEL 8.4 以下では modulemd-tools をインストールするのに Copr repository が必要
                # https://github.com/rpm-software-management/modulemd-tools
                verbose_run dnf -y copr enable frostyx/modulemd-tools-epel
                ;;
        esac
        PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL8
        PACKAGES_FOR_INSTALL="$PACKAGES_FOR_INSTALL yum python3-librepo rpm-plugin-*"
        PACKAGES_FOR_EXTRA="$PACKAGES_FOR_EXTRA createrepo_c modulemd-tools"
        repotrack_options="--destdir $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM --alldeps --resolve"
    fi

    if [ $SYSTEM == "AMZN" ]; then
        echo_info "Enable Amazon Linux Extra packages (epel/postgresql12)"
        PYTHON=$PLATFORM_PYTHON verbose_run amazon-linux-extras enable epel
        #
        # PG_YUMREPO は /etc/redhat-release が必要となるため、インストールできない。
        # Amazon Linux 標準の extra パッケージから postgresql-12 をインストールする
        #
        PYTHON=$PLATFORM_PYTHON verbose_run amazon-linux-extras enable "postgresql${PG_VERSION}"
    else
        echo_info "Install Yum repositories (EPEL/PGSQL)"
        install_repo_from_rpm "$EPEL_REPOPKG" "${EPEL_REPOPKG}*" $EPEL_YUMREPO
        install_repo_from_rpm "$PG_REPOPKG" "${PG_REPOPKG0}*" $PG_YUMREPO
    fi
    verbose_run yum-config-manager --enable $EPEL_REPONAME > /dev/null
    if [ $SYSTEM == "RHEL" ]; then
        verbose_run yum-config-manager --enable $OPTION_REPONAME > /dev/null
    fi
    if [ $SYSTEM != "AMZN" ]; then
        verbose_run yum-config-manager --enable $PG_REPONAME > /dev/null
    fi
    case $SYSTEM_NAME in
        "cent7" | "cent8" | "rhel7" | "rhel8" | "amzn")
            curl -s https://packagecloud.io/install/repositories/rabbitmq/erlang/script.rpm.sh | bash
            curl -s https://packagecloud.io/install/repositories/rabbitmq/rabbitmq-server/script.rpm.sh | bash
            verbose_run yum -y module disable "postgresql"
            # mod-wsgiをコンパイルするために必要
            PACKAGES_FOR_KOMPIRA="$PACKAGES_FOR_KOMPIRA redhat-rpm-config"
            PACKAGES_FOR_KOMPIRA="$PACKAGES_FOR_KOMPIRA $PACKAGES_FOR_LANGPACK"
            ;;
    esac
    if $EXTRA_WITH_HA; then
        if [ $SYSTEM_NAME == "cent8" ]; then
            verbose_run yum-config-manager --enable $CENT8_HA_REPONAME > /dev/null
        fi
    else
        PACKAGES_FOR_HA=
    fi

    echo_title "Create extra package"
    verbose_run mkdir -p $KOMPIRA_EXTRA_DIR/{/etc/yum.repos.d,$KOMPIRA_EXTRA_PIP,$KOMPIRA_EXTRA_RPM,$KOMPIRA_EXTRA_WHL}

    echo_info "Install rpm packages for create extra package"
    yum_install $PACKAGES_FOR_EXTRA $PACKAGES_FOR_SENDEVT_ONLINE $PACKAGES_FOR_JOBMNGR_ONLINE $PACKAGES_FOR_KOMPIRA_ONLINE

    echo_info "Download rpm packages"
    verbose_run repotrack $repotrack_options \
                $PACKAGES_FOR_INSTALL $PACKAGES_FOR_OFFLINE $PACKAGES_FOR_UPDATE \
                $PACKAGES_FOR_SENDEVT $PACKAGES_FOR_JOBMNGR $PACKAGES_FOR_KOMPIRA \
                $PACKAGES_FOR_KOMPIRA_RABBITMQ $PACKAGES_FOR_KOMPIRA_POSTGRESQL \
                $PACKAGES_FOR_SETUP_CLUSTER $PACKAGES_FOR_HA $WITH_RPM
    exit_if_failed "$?" "Failed to download rpm packages"
    # OS によっては repotrack gdb に失敗する（成功した場合だけ extra パッケージに含める）
    verbose_run repotrack $repotrack_options $PACKAGES_FOR_GDB
    # Amazon Linux 2 の場合、resource-agents-paf がレポジトリに含まれていないため、直接ダウンロードする
    if [ $SYSTEM == "AMZN" ]; then
        verbose_run curl -LO --output-dir $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM $RA_PKGURL
    fi
    if [ -n "$DEBUGINFO" ]; then
        echo_info "Download debuginfo packages"
        if [ $SYSTEM == "AMZN" ]; then
            verbose_run debuginfo-install -y --downloadonly --downloaddir=$KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM $DEBUGINFO
        else
            verbose_run debuginfo-install -y --disablerepo "pgdg*" --enablerepo "pgdg$PG_VERSION" --downloadonly --downloaddir=$KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM $DEBUGINFO
        fi
        exit_if_failed "$?" "Failed to download debuginfo packages"
    fi
    createrepo $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM
    if [ $RHEL_VERSION == 8 ]; then
        verbose_run createrepo_c $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM
        verbose_run repo2module -s stable $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM modules.yaml
        verbose_run modifyrepo_c --mdtype=modules modules.yaml $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM/repodata/
    fi
    cat > $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_REPO <<EOF
[$KOMPIRA_EXTRA_NAME]
name=$KOMPIRA_EXTRA_NAME
baseurl=file://$KOMPIRA_EXTRA_RPM
gpgcheck=0
enabled=0
EOF

    verbose_run $PYTHON -m venv .local
    verbose_run .local/bin/pip install -U pip wheel setuptools

    echo_info "Download wheel packages"
    verbose_run .local/bin/pip wheel -w $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_WHL \
                -r $requirements \
                "pip" "wheel" \
                $REQUIREMENTS_FOR_OFFLINE $WITH_WHL
    exit_if_failed "$?" "Failed to create wheelhouse"

    echo_info "Packaging extra package"
    verbose_run tar zcvf $KOMPIRA_EXTRA_DIR.tar.gz -C $KOMPIRA_EXTRA_DIR etc opt
}

usage_exit()
{
    cat <<EOF
Usage: $(basename $0) [OPTIONS] <kompira-server>

  --force                       Force install.
  --https                       Configure apache to force secure access (default).
  --no-https                    Configure apache to enable unsecure access.
  --initdata                    Initialize kompira database.
  --initfile                    Initialize $KOMPIRA_VAR_DIR/{repository}.
  --secret-key <secret-key>     Secret key for encrypted fields of DB.
  --no-backup                   Skip database backup.
  --rhui                        Install in RHUI mode.
  --rhel-option-repo <repo>     Specify RHEL repository.
  --skip-python3-install        Skip python3 installation.
  --skip-cluster-start          Skip cluster start.
  --locale-lang <lang>          Configure system locale laguage.
  --locale-timezone <zonename>  Configure system locale timezone.
  --proxy <proxy>               Specify a proxy in the form
                                [user:passwd@]proxy.server:port.
  --temp-proxy <proxy>          Proxy settings applied only during installation.
  --noproxy <hosts>             Comma-separated list of hosts which do not use proxy.
  --temp-noproxy <hosts>        No proxy settings applied only during installation.
  --jobmngr                     Install Kompira job manager package.
  --sendevt                     Install Kompira send event package.
  --with-rpm <rpms>             Install with additional rpm packages.
  --with-whl <wheels>           Install with additional wheel packages.
  --with-gdb                    Install with gdb and debuginfo.
  --offline                     Install in offline mode using the kompira-extra package.
  --extra                       Create the Kompira kompira-extra package for offline install.
  --extra-without-ha            Create the Kompira kompira-extra package without the HA packages.
  --install-only                Installation only, each daemon is not started.
  --dry-run                     Check the parameters without actually installing.
  --help                        Show this message.

EOF
    exit 1
}

parse_options()
{
    SPECIFY_SERVER=false
    OPTIONS=`getopt -q -o '' -l https,no-https,force,initdata,initfile,secret-key:,no-backup,backup-process,update,rhui,rhel-option-repo:,skip-python3-install,skip-cluster-start,locale-lang:,locale-timezone:,temp-proxy:,temp-noproxy:,proxy:,noproxy:,jobmngr,sendevt,requires-only,with-rpm:,with-whl:,with-gdb,offline,extra,extra-without-ha,install-only,dry-run,help -- "$@"`
    [ $? != 0 ] && usage_exit
    eval set -- "$OPTIONS"
    while true
    do
        arg="$1"
        case "$arg" in
            --https) HTTPS_MODE=true ;;
            --no-https) HTTPS_MODE=false ;;
            --force) FORCE_MODE=true ;;
            --initdata) INITDATA_MODE=true ;;
            --initfile) INITFILE_MODE=true ;;
            --secret-key) DB_SECRET_KEY=$2; shift ;;
            --no-backup) BACKUP_MODE=false ;;
            --backup-process) BACKUP_PROCESS=true ;;
            --rhui) RHUI_MODE=true ;;
            --rhel-option-repo) RHEL_OPTION_REPO=$2; shift ;;
            --skip-python3-install) SKIP_PYTHON3_INSTALL=true ;;
            --skip-cluster-start) SKIP_CLUSTER_START=true ;;
            --locale-lang) CONFIG_LOCALE_LANG=true; LOCALE_LANG=$2; shift ;;
            --locale-timezone) CONFIG_LOCALE_TIMEZONE=true; LOCALE_TIMEZONE=$2; shift ;;
            --temp-proxy) TEMP_PROXY_URL=$(normalize_proxy_url $2); shift;;
            --temp-noproxy) TEMP_NO_PROXY=$2; shift ;;
            --proxy) PROXY_URL=$(normalize_proxy_url $2); TEMP_PROXY_URL=$PROXY_URL; shift ;;
            --noproxy) NO_PROXY=$2; TEMP_NO_PROXY=$NO_PROXY; shift ;;
            --jobmngr) JOBMNGR_MODE=true; SPECIFY_SERVER=true ;;
            --sendevt) SENDEVT_MODE=true; SPECIFY_SERVER=true ;;
            --requires-only) REQUIRES_ONLY=true ;;
            --with-rpm) WITH_RPM=$2; shift ;;
            --with-whl) WITH_WHL=$2; shift ;;
            --with-gdb) WITH_GDB=true ;;
            --offline) OFFLINE_MODE=true; SETUP_TYPE="$SETUP_TYPE (offline-mode)" ;;
            --extra) EXTRA_MODE=true; SETUP_TYPE="$SETUP_TYPE (extra-mode)" ;;
            --extra-without-ha) EXTRA_MODE=true; EXTRA_WITH_HA=false; SETUP_TYPE="$SETUP_TYPE (extra-mode w/o HA)" ;;
            --install-only) INSTALL_ONLY=true ;;
            --dry-run) DRY_RUN=true ;;
            --update)
                echo "option $arg is deprecated."
                ;;
            --) shift; break ;; # 引数は無視
            *) usage_exit ;;
        esac
        shift
    done

    if $JOBMNGR_MODE && $SENDEVT_MODE; then
        echo '--jobmngr and --sendevt options are exclusive.'
        echo
        exit 1
    fi

    if $SPECIFY_SERVER; then
        if [ -z "$1" ]; then
            echo 'must specify <kompira-server>'
            echo
            usage_exit
        fi
        KOMPIRA_SERVER="$1"
    else
        if [ -n "$1" ]; then
            echo "unknown argument \"$1\""
            echo
            usage_exit
        fi
    fi

    if [ -n "$DB_SECRET_KEY" ] && [ ${#DB_SECRET_KEY} -lt "$DB_SECRET_KEY_MIN_LEN" ]; then
        echo "A secret key should be at least $DB_SECRET_KEY_MIN_LEN characters"
        echo
        exit 1
    fi
}

################################################################

set -o pipefail
parse_options "$@"
exec > >(tee $INSTALL_LOG) 2>&1
create_tmpdir
start_setup

#
# OPTION_REPONAMEの設定
#
if $RHUI_MODE; then
    #
    # RHUIレポジトリ名の実際の名称を repolist から取得して切り出すことで、名称の差異を吸収する
    #  - Azure: rhui-rhel-${RHEL_VERSION}-server-rhui-optional-rpms
    #  - AWS: rhel-${RHEL_VERSION}7-server-rhui-optional-rpms
    #
    RHUI_OPTION_REPONAME=$((yum repolist all | grep $RHUI_OPTION_REPONAME | cut -d ' ' -f 1) 2> /dev/null)
    OPTION_REPONAME=${RHUI_OPTION_REPONAME%%/*}
else
    OPTION_REPONAME=$RHEL_OPTION_REPONAME
fi
if [ -v RHEL_OPTION_REPO ]; then
    OPTION_REPONAME=$RHEL_OPTION_REPO
fi

#
# 環境チェック
#
check_root
check_locale_lang
check_environ
check_cluster
check_python
check_update
check_license
if $DRY_RUN; then
    exit_setup
fi
install_requires_prepare
yum_clean

#
# extra パッケージの作成
#
if $EXTRA_MODE; then
    create_extra
    exit_setup
fi

#
# システム設定
#
configure_locale_lang
configure_locale_timezone

#
# kompira アカウント/ホームディレクトリを作成する
#
create_user
create_home

#
# データベースのバックアップ
#
backup_kompira_database

#
# 動作中の kompira を停止し、アンインストールする
#
stop_kompira
uninstall_kompira
uninstall_packages

#
# virtualenv で仮想環境を作成する
#
create_virtualenv

#
# virtualenv 環境をアクティブにしてインストールを実施する
#
activate_virtualenv
install_kompira
deactivate_virtualenv

final_check
exit_setup
